'' =================================================================================================
''
''   File....... jm_ht16k33_4x14segs.spin
''   Purpose.... 4x14 segment display driver (e.g., Adafruit HT16K33 backpack)
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (C) 2016-2021 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon@jonmcphalen.com
''   Started....
''   Updated.... 09 JUN 2017
''
''   {$P1}
''
'' =================================================================================================


con { fixed io pins }

  RX1  = 31                                                     ' programming / terminal
  TX1  = 30

  SDA1 = 29                                                     ' eeprom / i2c
  SCL1 = 28


con

  DISP_OFF    = segs#DISP_OFF
  DISP_ON     = segs#DISP_ON

  BLINK_OFF   = segs#BLINK_OFF
  BLINK_2HZ   = segs#BLINK_2HZ
  BLINK_1HZ   = segs#BLINK_1HZ
  BLINK_0x5HZ = segs#BLINK_0x5HZ

  DP_SEG = %01000000_00000000


obj

  segs : "jm_ht16k33"                                           ' HT16K33 segment controller


var


pub null

'' This is not a top-level object


pub start(device)

'' Starts i2c display device on EE bus pins
'' -- device is address, %000 to %111

  startx(SCL1, SDA1, device)


pub startx(scl, sda, device)

'' Starts i2c display device
'' -- scl is the device SCL pin
'' -- sda is the device SDA pin
'' -- device is address, %000 to %111

  segs.start(scl, sda, device)


pub present

  return segs.present


pub set_display(state, rate)

'' Update display control bits
'' -- state is off (0) or on (1)
'' -- rate is the blink rate (0..3)

  segs.set_display(state, rate)


pub display_on(state)

'' Sets display on or off

  segs.display_on(state)


pub set_blink(rate)

  segs.set_blink(rate)


pub set_brightness(b)

'' Set display brightness, 0..15

  segs.set_brightness(b)


pub clear(update)

'' Clears display buffer

  segs.clear(update)


pub refresh(n) | idx

'' Writes display buffer to device
'' -- limit to n digits/characters
'' -- n is 1..4

  segs.refresh(n)


pub put_buf(col, bits)

'' Updates display buffer position with bits
'' -- pos is 0..3
'' -- bits is 14-bit segments

  segs.put_buf(col, bits)


pub direct_write(col, bits)

'' Writes directly to display
'' -- buffer is updated as well
'' -- pos is 0..3
'' -- bits is 14-bit segments

  segs.direct_write(col, bits)


pub alpha_bits(ascii)

'' Convert alpha to 14-segment bit pattern
'' -- returns 0 if out-of-range

  if ((ascii > 31) or (ascii < 128))
    return Segs14[ascii-32]                                     ' maps start as #32 (space)


pub digit(d)

'' Converts decimal digit d to bit pattern
'' -- returns 0 if out-of-range

  if ((d => 0) and (d =< 9))
    return Segs14[d+16]                                         ' 16 is offset to digits


pub str(p_str) | last, col

'' Print string at p_str
'' -- if shorter than 4 characters, trailing spaces added

  last := (strsize(p_str) <# 4) - 1
  if (last < 0)
    return

  repeat col from 0 to 3
    if (col > last)
      segs.put_buf(col, 0)
    else
      segs.put_buf(col, alpha_bits(byte[p_str+col]))

  segs.refresh(4)


pub scroll_str(p_str, ms)

'' Scroll string through display
'' -- add trailing spaces to end on clear display

  ms *= clkfreq / 1000                                          ' convert millis to ticks

  repeat strsize(p_str)-3                                       ' scroll to end of string
    str(p_str++)                                                ' output 4 chars
    waitcnt(cnt + ms)                                           ' hold


pub bin(value) | col

  value &= %1111                                                ' truncate to 4 bits

  repeat col from 0 to 3
    if (value & |<(3-col))
      segs.put_buf(col, Segs14[1+16])
    else
      segs.put_buf(col, Segs14[0+16])

  segs.refresh(4)


pub dec(value) | sign, divisor, col, dig

  segs.clear(false)                                             ' clear display buffer

  value := -999 #> value <# 9_999                               ' limit range of value

  if (value < 0)                                                ' check for - sign
    sign := "-"
    -value
  else
    sign := 0

  if (value == 0)                                               ' zero is special case
    divisor := 1
    col := 3
  else
    divisor := 1000                                             ' setup for max width
    col := 0
    repeat                                                      ' correct divisor for value
      if (value => divisor)
        quit
      divisor /= 10
      col += 1

  if (sign)                                                     ' display sign if needed
    segs.put_buf(col-1, alpha_bits("-"))

  repeat while (col < 4)                                        ' put valu digits in buffer
    dig := value / divisor + "0"
    segs.put_buf(col++, alpha_bits(dig))
    value //= divisor
    divisor /= 10

  segs.refresh(4)                                               ' update display


pub hex(value) | col, dig

  value <<= 16

  repeat col from 0 to 3
    dig := (lookupz((value <-= 4) & $F : "0".."9", "A".."F"))
    segs.put_buf(col, alpha_bits(dig))

  segs.refresh(4)


pub set_dpoints(dpbits) | p_buf, col

  p_buf := segs.address                                         ' base address of segments buffer

  repeat col from 0 to 3
    if (dpbits & |<col)
      word[p_buf][col] |= DP_SEG
    else
      word[p_buf][col] &= !DP_SEG

  segs.refresh(4)


pub address

'' Returns hub address of display buffer

  return segs.address


dat { segment maps }

  Segs14        word    %0000000000000000                       '  space
                word    %0000000000000110                       '  !
                word    %0000001000100000                       '  "
                word    %0001001011001110                       '  #
                word    %0001001011101101                       '  $
                word    %0000110000100100                       '  %
                word    %0010001101011101                       '  &
                word    %0000010000000000                       '  '
                word    %0010010000000000                       '  (
                word    %0000100100000000                       '  )
                word    %0011111111000000                       '  *
                word    %0001001011000000                       '  +
                word    %0000100000000000                       '  ,
                word    %0000000011000000                       '  -
                word    %0000000000000000                       '  .
                word    %0000110000000000                       '  /
                word    %0000110000111111                       '  0
                word    %0000000000000110                       '  1
                word    %0000000011011011                       '  2
                word    %0000000011001111                       '  3
                word    %0000000011100110                       '  4
                word    %0000000011101101                       '  5
                word    %0000000011111101                       '  6
                word    %0000000000000111                       '  7
                word    %0000000011111111                       '  8
                word    %0000000011100111                       '  9
                word    %0001001000000000                       '  :
                word    %0000101000000000                       '  ;
                word    %0010010000000000                       '  <
                word    %0000000011001000                       '  =
                word    %0000100100000000                       '  >
                word    %0001000010000011                       '  ?
                word    %0000001010111011                       '  @
                word    %0000000011110111                       '  A
                word    %0001001010001111                       '  B
                word    %0000000000111001                       '  C
                word    %0001001000001111                       '  D
                word    %0000000011111001                       '  E
                word    %0000000001110001                       '  F
                word    %0000000010111101                       '  G
                word    %0000000011110110                       '  H
                word    %0001001000000000                       '  I
                word    %0000000000011110                       '  J
                word    %0010010001110000                       '  K
                word    %0000000000111000                       '  L
                word    %0000010100110110                       '  M
                word    %0010000100110110                       '  N
                word    %0000000000111111                       '  O
                word    %0000000011110011                       '  P
                word    %0010000000111111                       '  Q
                word    %0010000011110011                       '  R
                word    %0000000011101101                       '  S
                word    %0001001000000001                       '  T
                word    %0000000000111110                       '  U
                word    %0000110000110000                       '  V
                word    %0010100000110110                       '  W
                word    %0010110100000000                       '  X
                word    %0001010100000000                       '  Y
                word    %0000110000001001                       '  Z
                word    %0000000000111001                       '  [
                word    %0010000100000000                       '
                word    %0000000000001111                       '  ]
                word    %0000110000000011                       '  ^
                word    %0000000000001000                       '  _
                word    %0000000100000000                       '  `
                word    %0001000001011000                       '  a
                word    %0010000001111000                       '  b
                word    %0000000011011000                       '  c
                word    %0000100010001110                       '  d
                word    %0000100001011000                       '  e
                word    %0000000001110001                       '  f
                word    %0000010010001110                       '  g
                word    %0001000001110000                       '  h
                word    %0001000000000000                       '  i
                word    %0000000000001110                       '  j
                word    %0011011000000000                       '  k
                word    %0000000000110000                       '  l
                word    %0001000011010100                       '  m
                word    %0000000011010100                       '  n
                word    %0000000011011100                       '  o
                word    %0000000101110000                       '  p
                word    %0000010010000110                       '  q
                word    %0000000001010000                       '  r
                word    %0010000010001000                       '  s
                word    %0000000001111000                       '  t
                word    %0000000000011100                       '  u
                word    %0010000000000100                       '  v
                word    %0010100000010100                       '  w
                word    %0010100011000000                       '  x
                word    %0010000000001100                       '  y
                word    %0000100001001000                       '  z
                word    %0000100101001001                       '  {
                word    %0001001000000000                       '  |
                word    %0010010010001001                       '  }
                word    %0000010100100000                       '  ~
                word    %0011111111111111                       ' all (no dp)



con { license }

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}